// C Header File
// Created 01/01/03; 12:31:40

//--------------------------------------------------------------------------
// <sub_routines.h>
// Sub-routines used by main.c
//--------------------------------------------------------------------------

#ifndef __SUB_H__
#define __SUB_H__

#include "TIFSimport.h"
#include "macros.h"
#include "functions.h"
#include "structures.h"

//==========================================================================
// Creates a list of AppDataStruct for all the Flash Apps found in ROM. An
// AppsDataStruct contains a pointer to the name of the App, a pointer to
// the App in Flash ROM, if the App is not hidden, its AppID (H_NULL
// if the App is hidden), and field telling what the user wants to do with
// the App (filled and used by HideApps() ).
// The AppsDataStruct are sorted in alphabetical order for the names.
// PrepareAppsData() returns the handle of the list created (don't forget to
// free it after use), or H_NULL if there is not enough memory.
// If there is enough memory, the number of apps in the list created is
// stored in the variable pointed by ListSize.
// If the list is empty, returns 0 in this variable, and returns H_NULL.
// If there is not enough memory, returns x0FFFF in the variable.
//==========================================================================
static inline HANDLE PrepareAppsData (unsigned short *ListSizePtr)
{
	unsigned short ListSize;   // number of apps in the list
	HANDLE ListHandle;   // handle the list

	*ListSizePtr = 0xFFFF;   // if we find later that there is not enough memory

	//=====================================================================
	// Create the list of AppDataStruct and fill them
	//=====================================================================
	{
		AppDataStruct *ListPtr;   // pointer in the list
		CFILE context;
		CERT_FIELD cert_field;
		
		// get a pointer to the end of the AMS
		copen (&context, PRODUCT_MEMORY, (long)PRODUCT_LENGTH);
		cread (&context, &cert_field);
		cread (&context, &cert_field);
		// get a pointer to the first flash app
		context.Pos = (void*) ((long)(context.Pos - 1 + 0x10000) & 0xFFFF0000);
		// read the first app if any
		cread (&context, &cert_field);
		
		ListSize = 0;
		if (!(ListHandle = HeapAlloc (2)))  // just to create a block of memory
			return H_NULL;
		
		// search all the Apps in ROM
		while (cert_field.Field != NO_MORE_APPS_ID)
		{
				void *NewAppPtr;
			
			ListSize++;   // one more app
			// place for a new pointer in the list
			if (!HeapRealloc (ListHandle, ListSize * sizeof (AppDataStruct))) {
				HeapFree (ListHandle);
				return H_NULL;   // no memory
			}
			// Get a pointer to this app (at an even address boundary)
			NewAppPtr = (void*) ((long)(cert_field.Data - cert_field.HdrLen + 1) & ~1);			
			// add the pointer to the app to the list
			ListPtr = &((AppDataStruct*)HeapDeref (ListHandle))[ListSize - 1];
			ListPtr->AppPtr = NewAppPtr;
   		// write its AppID in the list (H_NULL if not found -> app hidden)
	 		ListPtr->ID = GetAppIDFromAppPtr (NewAppPtr);
			
			// Get the name of the App
			GetFlashAppName (ListPtr);
			// go to the next app (if any)
			do cread (&context, &cert_field);
			while (cert_field.Field != NO_MORE_APPS_ID && cert_field.Field != NEW_APP_ID);

		}   // 	while (cert_field.Field ...
		
	}   // CreateAppsDataList block
	
	
	//=====================================================================
	// Sort the list in alphabetical order of the  Apps names (only if
	// there are apps to sort).
	//=====================================================================
	if (ListSize > 1)
	{
		BOOL swapped = TRUE;
		unsigned short i, j;
		AppDataStruct *AppsListPtr = HeapDeref (ListHandle), *AppsListScanPtr;
		void* TempPtr;
		AppID TempID;
	
		// We use the bubble sort
		for (i = ListSize - 1; i >= 2 && swapped; i--) {
			swapped = FALSE;
			for (j = 1, AppsListScanPtr = AppsListPtr; j <= i; j++, AppsListScanPtr++) {
				// compare the names of 2 apps (without taking into account the case)
				if (stricmp (AppsListScanPtr->name, (AppsListScanPtr + 1)->name) > 0) {
					// swap the content of the blocks
					TempPtr =  AppsListScanPtr->name;
					AppsListScanPtr->name = (AppsListScanPtr + 1)->name;
					(AppsListScanPtr + 1)->name = TempPtr;
					TempPtr =  AppsListScanPtr->AppPtr;
					AppsListScanPtr->AppPtr = (AppsListScanPtr + 1)->AppPtr;
					(AppsListScanPtr + 1)->AppPtr = TempPtr;
					TempID =  AppsListScanPtr->ID;
					AppsListScanPtr->ID = (AppsListScanPtr + 1)->ID;
					(AppsListScanPtr + 1)->ID = TempID;
					swapped = TRUE;
				}   // if (stricmp ( ...

			}   // for (j = 0, ...
		}   // for (i = ListSize - 2; ...
			
	}   // 	if (ListSize > 1)

	if (!(*ListSizePtr = ListSize)) {   // return the size
		HeapFree (ListHandle);   // no apps in the list
		return H_NULL;
	}
	return ListHandle;   // and the handle

}   // PrepareAppsData ()



//==========================================================================
// Creates and show a dialog window to select the Apps to hide or unhide.
// ListSize and AppsData_handle are the parameters of the list of
// AppDataStruct returned by PrepareAppsData.
// Returns the handle of a block of memory containing the result of the
// selection (an array of unsigned short, one for each App, containing
// 1 (option 1 : do nothing) or 2 (option 2 : hide/unhide) ).
// Returns H_NULL if there is not enoureturnsgh memory.
// If there was enough memory,  returns TRUE in the variable pointed by
// EnterPressed if 'Enter' was pressed, FALSE if 'Esc' was pressed.
//==========================================================================
static inline HANDLE ShowMenu (unsigned short ListSize,
	HANDLE AppsData_handle, BOOL *EnterPressed)
{
	HANDLE ResultsHandle, DialogHandle;
	HANDLE *MenuHandlesPtr;   // points to the array of handles of the dropdowns^
	unsigned short LastFieldBottom_y;   // y coordinate after the last field
	unsigned short x_coord;   // x coordinate of all the fields created
	
	x_coord = (ListSize > MENU_MAX_PULLDOWNS) ? MENU_COORD_X : MENU_COORD_X_NO_SCROLLING;
	LastFieldBottom_y = PULLDOWN_Y_ORIGIN + min (ListSize, MENU_MAX_PULLDOWNS) \
	 * PULLDOWN_Y_STEP;
	// A block of memory where DialogDo() will write the results of the
	// pulldowns.
	if (!( ResultsHandle = HeapAlloc (ListSize * sizeof (short)) ))
		return H_NULL;
	// the initial option of the dropdowns will be the first one
	memset (HeapDeref (ResultsHandle), 1, ListSize * sizeof (short));
	// new dialog
	if (!( DialogHandle = DialogNewSimple (MENU_W, LastFieldBottom_y + MENU_BOTTOM) )) {
		HeapFree (ResultsHandle);
		return H_NULL;
	}
	
	// add a scroll region (must be first !)
	if (!DialogAdd (DialogHandle, DF_CLR_ON_REDRAW, x_coord, PULLDOWN_Y_ORIGIN,
	 // WORD x1, WORD y1, WORD Index0, WORD Index1, WORD NumDspFields,
	 // WORD TotNumFields, WORD FieldHeight
	 // Don't forget there is one more field on TI89
	 D_SCROLL_REGION, MENU_W, LastFieldBottom_y, CST_C (3,2),
	 CST_C (3,2) + ListSize , MENU_MAX_PULLDOWNS, ListSize , \
	 PULLDOWN_Y_STEP) ||
	// add the title
	!DialogAddTitle (DialogHandle, MENU_TITLE,  BT_CANCEL, BT_OK)) {
	 	HeapFree (DialogHandle);
		HeapFree (ResultsHandle);
		return H_NULL;
	}   // if (!DialogAdd ( ...
	
	// add the help text on TI89
	if (TI89)
		if (!DialogAddText (DialogHandle, MENU_TEXT_X, MENU_TEXT_Y, MENU_TEXT_89)) {
	 	HeapFree (DialogHandle);
		HeapFree (ResultsHandle);
		return H_NULL;
	}			
	
	// create a block of memory for the handles of the popups (for the dropdowns)
  // Set all the handles to H_NULL.
	if (!( MenuHandlesPtr = calloc (ListSize, sizeof (HANDLE)) )) {
		HeapFree (DialogHandle);
		HeapFree (ResultsHandle);
		return H_NULL;		
	}
	
	
	//=====================================================================
	// Create a popup for each App
	//=====================================================================
	{
		AppDataStruct *AppsDataRead_ptr, // pointer in the array of AppsDataStruct
			*AppsDataPtr;   
		unsigned short index;  // index in the array where the results are written
		unsigned coord_y;   // y coordinate of the current pulldown
		HANDLE MenuHandle;
		HANDLE *MenuHandlesFill_ptr;   // pointer in the array MenuHandlesPtr[]
	
		// dereference it and lock it (we will allocate other blocks of memory)
		if (!(AppsDataRead_ptr = AppsDataPtr = HLock (AppsData_handle))) {
			 	// not enough memory
			 	DestroyMenuHandles (MenuHandlesPtr, ListSize);
				HeapFree (DialogHandle);
				HeapFree (ResultsHandle);
				return H_NULL;				
		}   // if (!(AppsDataRead_ptr ...
			
		coord_y = PULLDOWN_Y_ORIGIN;
		// for all the Apps (AppsDataRead_ptr++ to get the AppDataStruct of the next
		// App)
		for (index = 0, MenuHandlesFill_ptr = MenuHandlesPtr; index < ListSize;
		index++, coord_y += PULLDOWN_Y_STEP, AppsDataRead_ptr++) 
		
		{
		char AppName[APP_NAME_MAX_LEN + sizeof (HIDDEN_SYMBOL_STR)];
	
		// put HIDDEN_SYMBOL_STR if the App is hidden
		strcpy (AppName, !(AppsDataRead_ptr->ID) ? HIDDEN_SYMBOL_STR : \
			NO_H_SYMBOL_STR);	
		strcat (AppName, AppsDataRead_ptr->name);
		
		if (
			!( MenuHandle = *MenuHandlesFill_ptr++ = PopupNew (NULL, 0)) ||	
			// first option : nothing (do not change the state of the App)
			!PopupAddText (MenuHandle, -1, "", 0) ||
			// second option : "Hide" if the App is not hidden (ie it has an ACB)
			!PopupAddText (MenuHandle, -1, AppsDataRead_ptr->ID ? "Hide" : "Unhide", 0) ||
			// index is the offset of the result in the array of handle ResultsHandle.
			// The prompt is the name of the App.
			// It's in a scroll region, and we want a tabulation between the label and
			// the popup.
			!DialogAddPulldownAndFlags (DialogHandle, DF_SCROLLABLE | DF_TAB_SPACES, \
			 x_coord, coord_y, AppName, MenuHandle, index) )
			 
			 {
			 	// not enough memory
			 	DestroyMenuHandles (MenuHandlesPtr, ListSize);
				HeapFree (DialogHandle);
				HeapFree (ResultsHandle);
				return H_NULL;	
			}   // 	if (!( MenuHandle = ...
			
		}   // for (index = 0; ...
			
	}   // CreatePopups block
	
	
	//=====================================================================	
	// show the dialog box
	//===================================================================== 
	*EnterPressed = 
	DialogDo (DialogHandle, CENTER, CENTER, NULL, \
		HeapDeref (ResultsHandle)) == KEY_ENTER;
	DestroyMenuHandles (MenuHandlesPtr, ListSize);
	HeapFree (DialogHandle);
	return ResultsHandle;
	
}   // ShowMenu ()



//==========================================================================
// Hides and unhide the Flash Apps. AppsData_handle is the handle of the list
// of AppDataStruct created by PrepareAppsData(). MenuResults_handle is the
// handle of the block containing the selected options of the menu. ListSize
// is the number of Apps in ROM.
// SomethingDone is a pointer to a variable where TRUE will be returned if
// one or more Apps were hidden or unhidden.
// HideApps() returns FALSE if some Apps could not be hidden/unhidden.
// if *SomethingDone = FALSE, HideApps() returns TRUE.
//==========================================================================
static inline BOOL HideApps (HANDLE MenuResults_handle,
	HANDLE AppsData_handle,	unsigned short ListSize, BOOL *SomethingDone)
{
	unsigned short *MenuResultsPtr;   // points to the results of the menu
	unsigned short *MenuResultsRead_ptr; // points in the results of the menu
	AppDataStruct *AppsDataPtr;   // points to the array of AppsDataStruct
	AppDataStruct *AppsDataRead_ptr;   // points in the array of AppsDataStuct
	EVENT event;
	BOOL CouldNotChange = FALSE;   // TRUE if some Apps could not be hidden/unhidden

	if (!(MenuResultsPtr = HLock (MenuResults_handle)) || \
	 !(AppsDataPtr = HLock (AppsData_handle))) {
		// not enough memory
		*SomethingDone = FALSE;
		return TRUE;
	}
	
	//=====================================================================
	// Hide the Apps to hide. Search if there are Apps to hide or unhide.
	// Search if there are Apps to unhide. Fill the element 'action' of
	// the AppDataStructs.
	//=====================================================================
	{
		BOOL SomethingToDo = FALSE, AppsToHide = FALSE;
		unsigned short i;
		AppID ID;
		
		// for all the Apps in ROM
		AppsDataRead_ptr = AppsDataPtr;
		MenuResultsRead_ptr = MenuResultsPtr;
		for (i = ListSize; i; i--, AppsDataRead_ptr++){
		
			if (*MenuResultsRead_ptr++  != 2) {
				AppsDataRead_ptr->action = APA_NONE;
				continue;
			}
				
			// second option selected : change the state of the App
			ID = AppsDataRead_ptr->ID;
			SomethingToDo = TRUE;
			if (ID) {
				
				// the App is installed : hide it.
				AppsDataRead_ptr->action = APA_HIDE;	
				// Can we delete the App ? (thanks to Greg Dietsche for telling
				// me to add this test)
				if (!AppCanDelete (ID)) {
					CouldNotChange = TRUE;
					continue;   // no : deal with the next App
				}
					
				// Send CM_UNINSTALL to it to let the App know it must free
				// the blocks of memory it has created.
				event.Type = CM_UNINSTALL;
				EV_sendEvent (ID, &event);
				// remove its AppID from the linked list of the ACBs
				RemoveACB (ID);
				// free its object frame chain
				OO_DestroyAll (((ACB*)HeapDeref (ID))->appData);
				// free its ACB
				HeapFree (ID);
				
			}
			
			else {  // -> if (!ID)
				// the App is hidden. We will unhide it later.
				AppsDataRead_ptr->action = APA_UNHIDE;	
				AppsToHide = TRUE;
			}   // -> if (!ID)
					
		}   // for (i = ListSize; ...
		
		*SomethingDone = SomethingToDo;
		if (!AppsToHide)
			return (!CouldNotChange);   // we had only Apps to unhide	
		
	}   // Hide block
	
	
	//=====================================================================
	// Now unhide the Apps to unhide.
	//=====================================================================
	
	{
		unsigned short i;
		
		ST_helpMsg ("Unhiding Apps...");
		// We cannot Unrelocate only one App :( Unrelocate all of them.
		// Note that this unrelocation is why unhidding Apps is so slow.
		PrepareAppsMem (TRUE, FALSE);

		// Remove =from the linked list the ACBs of the Apps not to hide or
		// unhide and that have an ACB, NOW !
		// These ACBs are for unrelocated Apps. There will be a problem when
		// InstallApp() is run further if we do not do this now (It was very
		// difficult to find the problem : the HashTable for the Apps attributes
		// contained invalid data written during the sort-by-name of the Apps
		// made by InstallApp(), etc. :( )
		AppsDataRead_ptr = AppsDataPtr;
		for (i = ListSize; i; i--, AppsDataRead_ptr++) {
			if (AppsDataRead_ptr->action == APA_NONE && AppsDataRead_ptr->ID)
			// remove the ACBs from the linked list (but keep the block, it
			// contains BSS that we will copy to the new ACB)
			RemoveACB (AppsDataRead_ptr->ID);
		}   // 	for (i = ListSize; ...
		
		
		// Relocate all the Apps
		AppsDataRead_ptr = AppsDataPtr;
		for (i = ListSize; i; i--, AppsDataRead_ptr++)
		{
			AppID ID;
			AppID NewID;
			
			ID = AppsDataRead_ptr->ID;
			// Relocate the App
			if (!InstallApp (AppsDataRead_ptr->AppPtr)) {
				CouldNotChange = TRUE;   // not enouh memory for this App
				// if the App is not hidden
				if (AppsDataRead_ptr->action == APA_NONE && ID) {
					// the old ACB still exists !
					HeapFree (ID);
				}
				// else it was a hidden App (without ACB)
				continue;   // deal with the next App
			}   // if (!InstallApp ( ...
			
			NewID = GetAppIDFromAppPtr (AppsDataRead_ptr->AppPtr);
				
			if (AppsDataRead_ptr->action == APA_NONE && ID)
			{
			 	ACB *NewACBPtr = HeapDeref (NewID);
				ACB *OldACBPtr =  HeapDeref (ID);

				// We must copy the BSS in the old ACB to the new ACB.
				// Do NOT copy the SACB header of the ACB.
				memcpy ((char*)NewACBPtr + sizeof (struct SACB), \
				 (char*)OldACBPtr + sizeof (struct SACB), \
				 HeapSize (ID) - sizeof (struct SACB) );
				// keep the same flags, except ACB_INSTALLED (SendInstallMsgs()
				// will need it cleared)
				NewACBPtr->flags = OldACBPtr->flags;
				// free the old ACB
				HeapFree (ID);
			}   // if (AppsDataRead_ptr->action ...
			
			else if (AppsDataRead_ptr->action == APA_HIDE || 
			 (AppsDataRead_ptr->action == APA_NONE && !ID)) {
				// we had to relocate it (because else it first word would not be
				// anymore REAL_FLASH_APP_ID), but we do not need the ACB created.
				// Remove it from the linked list, and free it.
				RemoveACB (NewID);
				HeapFree (NewID);
			}
			
		}   // for (i = ListSize; ...
		
		// The Apps to unhide are really reinstalled. We must send CM_INSTALL
		// to them, and call the method OO_APP_NOTICE_INSTALL of the other Apps.
		// This will be done for all the Apps to unhide (there flags ACB_INSTALLED
		// has not been set yet)
		SendInstallMsgs();
		
		ST_eraseHelp ();
		return (!CouldNotChange);
		
	}   // Unhide block
	
}   // HideApps ()



//==========================================================================

//==========================================================================
static inline void ShowMemState (unsigned long OldMemState)
{
	char sbuffer[sizeof ("Memory gain : 999999 bytes")];
	long MemGain;
	
	HeapCompress ();
	MemGain  = HeapAvail () - OldMemState;
	strcpy (sbuffer, "Memory ");
	strcat (sbuffer, (MemGain >= 0) ? "gain : " : "loss : ");
	sprintf (&sbuffer[sizeof ("Memory gain : ") - 1], "%lu", abs (MemGain));
	strcat (sbuffer, " bytes");
	
	DlgMessage (MEM_MENU_TITLE, sbuffer, BT_NONE, BT_OK);
	
}   // ShowMemState ()



//-------------------- Internal Sub-Routines for sub.h ---------------------

//==========================================================================
// Fills AppData with the name of the App. The other elements of AppData
// must have been filled. If ID is null, its internal name is searched (8
// characters at the most). Else its full name is searched (20/32 characters
// at the most).
//==========================================================================
static inline void GetFlashAppName (AppDataStruct *AppData)
{
	long *AppPtr;

	if (AppData->ID) {
		AppData->name = GetAppName (AppData->ID);
		return;
	}
	
	// We haven't any pointer to the App Header. We search the magic number
	// that marks its beginning. Perhaps a bit nasty...
	AppPtr = AppData->AppPtr;
	while (*AppPtr != APP_HEADER_MAGIC)
		((short*)AppPtr)++;
	// the internal App name is just after the magic number
	AppData->name = (char*)++AppPtr;

}   // GetFlashAppName ()



//==========================================================================
// Frees all the handles (!= H_NULL) of array MenuHandlesPtr[] of size
// ListSize.
//==========================================================================
void DestroyMenuHandles (HANDLE * const MenuHandlesPtr,
	unsigned short ListSize) 
{
	HANDLE *MenuHandlesScan_ptr = MenuHandlesPtr;

	for ( ; ListSize; ListSize--)
		HeapFree (*MenuHandlesScan_ptr++);
	HeapFreePtr (MenuHandlesPtr);   // and free the array
}   // DestroyMenuHandles ()



//==========================================================================
// Returns the AppID of an App, giving a pointer to it. If the App has no
// ACB, returns H_NULL.
//==========================================================================
AppID GetAppIDFromAppPtr (void *AppPtr)
{
	AppID CurrentACB = OO_firstACB;
	ACB *ACBPtr;

	// scan the linked list of the ACBs
	while (CurrentACB) {
		ACBPtr = HeapDeref (CurrentACB);
		if (ACBPtr->certhdr == AppPtr)
			break;   // the app is not hidden
		CurrentACB = ACBPtr->next;
	}   // while (CurrentACB)
	return CurrentACB;
	
}   // GetAppIDFromAppPtr ()



//==========================================================================
// Removes an ACB from the linked list of the OS.
//==========================================================================
void RemoveACB (AppID ID)
{
	ACB *ACBPtr = HeapDeref (ID);
	if (ACBPtr->prev)
 		((ACB*)HeapDeref (ACBPtr->prev)) -> next = ACBPtr->next;
 	if (ACBPtr->next)
 		((ACB*)HeapDeref (ACBPtr->next)) -> prev = ACBPtr->prev;
}


#endif     // #ifndef __SUB_H__
